#! /usr/local/bin/python


"""
usage: freeze [options] script [module] ...

The options are:

-e extension:   A directory containing additional .o files that
                may be used to resolve modules.  This directory
                should also have a Setup file describing the .o files.
                More than one -e option may be given.

-m:             Invoke 'make' afterwards.

-O:             Optimize.

-p prefix:      This is the prefix used when you ran ``make install''
                in the Python build directory. 
                (If you never ran this, freeze won't work.)
                The default is whatever sys.prefix evaluates to.

-P exec_prefix: Like -p but this is the 'exec_prefix', used to
                install objects etc.  The default is whatever
		sys.exec_prefix evaluates to, or the -p argument if given.

-o path:        This names a directory where the target will by build.
                The directory will be created if not already present.
                When used together with -m, the directory will be
                deleted after a succesful 'make' (but only if it was
                created int this run of freeze). Defaults to '.'

-i modulename:  Builtin to add to modulelist

-n passes:	Number of optimization passes to make, defaults to 2.

-s		Directs the optimizer to strips line numbers from the frozen 
		byte code.

-d		Directs the optimizer to not strip doc strings.

script:         The Python script to be executed by the resulting binary.
                It *must* end with a .py suffix!

module ...:     Additional Python modules (referenced by pathname)
                that will be included in the resulting binary.  These
                may be .py or .pyc files.

NOTES:

In order to use freeze successfully, you must have built Python and
installed it ("make install").
"""


# Import standard modules

import cmp
import getopt
import os
import string
import sys
import addpack

# Import the freeze-private modules

import checkextensions
import findmodules
import makeconfig
import makefreeze
import makemakefile
import parsesetup


# Main program

# These must be global because of the exception handler around main
odir = ''				# settable with -t option
created_odir = None

def main():
	global odir, created_odir
	# overridable context
	runmake = None			# settable with -m option
	optimize = 0			# settable with -O option
	prefix = None			# settable with -p option
	exec_prefix = None		# settable with -P option
	extensions = []
        excludelist = []		# settable with -x option
        includelist = []		# settable with -i option
	passes	= 2			# settable with -n option
	strip = None			# settable with -s option
	stripdoc = 1			# settable with -d option
	path = sys.path

	# output files
	frozen_c = 'frozen.c'
	config_c = 'config.c'
	target = 'a.out'		# normally derived from script name
	makefile = 'Makefile'

	# parse command line
	try:
		opts, args = getopt.getopt(sys.argv[1:], 'e:mOsp:P:t:x:i:o:n:d')
	except getopt.error, msg:
		usage('getopt error: ' + str(msg))

	# proces option arguments
	for o, a in opts:
		if o == '-e':
			extensions.append(a)
		if o == '-m':
			runmake = 1
		if o == '-O':
			optimize = 1
		if o == '-p':
			prefix = a
		if o == '-P':
			exec_prefix = a
		if o == '-o':
			odir = a
                if o == '-x':
                        excludelist.append(a)
                if o == '-i':
                        includelist.append(a)
		if o == '-n':
			passes=string.atoi(a)
		if o == '-s':
			strip=1
		if o == '-d':
			stripdoc=0

	# default exec_prefix and exec_prefix
	if not exec_prefix:
		if prefix:
			exec_prefix = prefix
		else:
			exec_prefix = sys.exec_prefix
	if not prefix:
		prefix = sys.prefix

	# locations derived from options
	version = sys.version[:3]
	binlib = os.path.join(exec_prefix, 'lib/python%s/config' % version)
	incldir = os.path.join(prefix, 'include/python%s' % version)
	config_c_in = os.path.join(binlib, 'config.c.in')
	frozenmain_c = os.path.join(binlib, 'frozenmain.c')
	getpath_c = os.path.join(binlib, 'getpath.c')
	supp_sources = [frozenmain_c, getpath_c]
	makefile_in = os.path.join(binlib, 'Makefile')
	defines = ['-DPYTHONPATH=\\"$(PYTHONPATH)\\"']
	includes = ['-I' + incldir, '-I' + binlib]

	# sanity check of directories and files
	for dir in [prefix, exec_prefix, binlib, incldir] + extensions:
		if not os.path.exists(dir):
			usage('needed directory %s not found' % dir)
		if not os.path.isdir(dir):
			usage('%s: not a directory' % dir)
	for file in [config_c_in, makefile_in] + supp_sources:
		if not os.path.exists(file):
			usage('needed file %s not found' % file)
		if not os.path.isfile(file):
			usage('%s: not a plain file' % file)
	for dir in extensions:
		setup = os.path.join(dir, 'Setup')
		if not os.path.exists(setup):
			usage('needed file %s not found' % setup)
		if not os.path.isfile(setup):
			usage('%s: not a plain file' % setup)

	# check that enough arguments are passed
	if not args:
		usage('at least one filename argument required')

	# check that the script name ends in ".py"
	if args[0][-3:] != ".py":
		usage('the script name must have a .py suffix')

	# check that file arguments exist
	for arg in args:
		if not os.path.exists(arg):
			usage('argument %s not found' % arg)
		if not os.path.isfile(arg):
			usage('%s: not a plain file' % arg)

	# process non-option arguments
	scriptfile = args[0]
        modules = args[1:]

	# derive target name from script name
	base = os.path.basename(scriptfile)
	base, ext = os.path.splitext(base)
	if base:
		if base != scriptfile:
			target = base
		else:
			target = base + '.bin'

	# handle -o option
	base_frozen_c = frozen_c
	base_config_c = config_c
	base_target = target
	if odir and not os.path.isdir(odir):
		try:
			os.mkdir(odir)
			print "Created output directory", odir
		except os.error, msg:
			usage('%s: mkdir failed (%s)' % (odir, str(msg)))
	if odir:
		frozen_c = os.path.join(odir, frozen_c)
		config_c = os.path.join(odir, config_c)
		makefile = os.path.join(odir,makefile)

	# Actual work starts here...

	dict = findmodules.findmodules(scriptfile, modules, path,excludelist)
	names = dict.keys()
	names.sort()

	backup = frozen_c + '~'
	try:
		os.rename(frozen_c, backup)
	except os.error:
		backup = None
	outfp = open(frozen_c, 'w')
	try:
		makefreeze.makefreeze(outfp, dict, optimize, strip, passes,
stripdoc)
	finally:
		outfp.close()
	if backup:
		if cmp.cmp(backup, frozen_c):
			sys.stderr.write('%s not changed, not written\n' %
					 frozen_c)
			os.rename(backup, frozen_c)

	builtins = []
	unknown = []
	mods = dict.keys()
	mods.sort()
	for mod in mods:
		if dict[mod] == '<builtin>':
			builtins.append(mod)
		elif dict[mod] == '<unknown>':
			unknown.append(mod)

	addfiles = []
	if unknown:
		addfiles, addmods = \
			  checkextensions.checkextensions(unknown, extensions)
		for mod in addmods:
			unknown.remove(mod)
		builtins = builtins + addmods
	if unknown:
		sys.stderr.write('Warning: unknown modules remain: %s\n' %
				 string.join(unknown))
        for blah in includelist:
           builtins.append(blah)
	builtins.sort()
	infp = open(config_c_in)
	backup = config_c + '~'
	try:
		os.rename(config_c, backup)
	except os.error:
		backup = None
	outfp = open(config_c, 'w')
	try:
		makeconfig.makeconfig(infp, outfp, builtins)
	finally:
		outfp.close()
	infp.close()
	if backup:
		if cmp.cmp(backup, config_c):
			sys.stderr.write('%s not changed, not written\n' %
					 config_c)
			os.rename(backup, config_c)

	cflags = defines + includes + ['$(OPT)']
	libs = []
	for n in 'Modules', 'Python', 'Objects', 'Parser':
		n = 'lib%s.a' % n
		n = os.path.join(binlib, n)
		libs.append(n)

	makevars = parsesetup.getmakevars(makefile_in)
	somevars = {}
	for key in makevars.keys():
		somevars[key] = makevars[key]

	somevars['CFLAGS'] = string.join(cflags) # override
	files = ['$(OPT)', '$(LDFLAGS)', base_config_c, base_frozen_c] + \
		supp_sources +  addfiles + libs + \
		['$(MODLIBS)', '$(LIBS)', '$(SYSLIBS)']

	backup = makefile + '~'
	try:
		os.rename(makefile, backup)
	except os.error:
		backup = None
	outfp = open(makefile, 'w')
	try:
		makemakefile.makemakefile(outfp, somevars, files, target)
	finally:
		outfp.close()
	if backup:
		if not cmp.cmp(backup, makefile):
			print 'previous Makefile saved as', backup
		else:
			sys.stderr.write('%s not changed, not written\n' %
					 makefile)
			os.rename(backup, makefile)

	# Ready for make!

	if not odir:
		if runmake:
			print 'running make'
			status = os.system('make > /dev/null')
			if status:
				sys.stderr.write('Compilation failed\n')
				sys.exit(1)
		else:
			print 'Now run make to build the target:', base_target
	elif runmake:
		# Run make in the odir, remove the directory altogether
		# if this process created it, and the make was succesfull.
		curdir = os.getcwd()
		os.chdir(odir)
		print 'running make'
		status = os.system('make > /dev/null')
		os.chdir(curdir)

		if status:
			sys.stderr.write(
				'Compilation failed. Files left in %s\n' % 
				 (odir,))
			sys.exit(1)
		else:
			tmptarget = os.path.join(odir, target)
			try: os.rename(tmptarget, target)
			except os.error:
				os.system('cp %s %s' % (tmptarget, target))
			if created_odir:
				os.system('rm -rf %s' % (odir,))
				odir = None
			print 'Frozen target:', target


# Print usage message and exit

def usage(msg = None):
	sys.stderr.write(__doc__)
	# Put the error last since the usage message scrolls off the screen
	if msg:
		sys.stderr.write(str(msg) + '\n')
	sys.exit(2)


# This could trap a ^c or such.
#try:
main()
#except:
#	if odir and created_odir:
#		print 'cleaning up'
#		os.system('rm -rf ' + odir)
