#!/usr/local/bin/python
# Optimise a code object

rcs_ident = '$Header: /var/home/siebren/src/xfreeze/RCS/optimize.py,v 1.12 1995/10/20 08:40:46 siebren Exp $'

import new, sys
import types

import opcode
import ltup
from revmap import ReviseMap
from ltup import abs_jump, rel_jump, cnd_jump, unc_jump, unc_jumpr, \
	name_instructions

numericTypes = (types.IntType, types.LongType, types.FloatType, types.ComplexType)
intAndStrTypes = (types.IntType, types.LongType, types.StringType)

class Optimising:
	def __init__(self, code, strip_setlineno, parentOpt, stripdoc = 1):
		self.oldcode = code
		self.strip_setlineno = strip_setlineno
		self.parentOpt = parentOpt
		self.statistics = {}
		self.the_consts = ReviseMap(code.co_consts)
		self.the_names = ReviseMap(code.co_names)
		self.stripdoc = stripdoc


	# Statistics are gathered by self.did(string),
	# and printed by self.stat().
	# Statistics for code objects found recursively
	# are added by self.statadd
	def did(self, thing, amount = 1):
		'increment one of the statistics counters'
		if type(thing) <> types.StringType:
			raise TypeError, 'Optimizing.did expects a string'
		if amount:
			dict = self.statistics
			if dict.has_key(thing):
				dict[thing] = dict[thing] + amount
			else:
				dict[thing] = amount
	# end did


	def statadd(self, things):
		if type(things) <> types.DictType:
			raise TypeError, 'Optimizing.statadd expects a dict'
		for k in things.keys():
			self.did(k, things[k])
	# end statadd


	def stat(self, terminator = '\n'):
		'Return a multiline statistics representation'
		d = self.statistics
		keys = d.keys()
		keys.sort()
		ret = ''
		for k in keys:
			ret = ret + repr(d[k]) + ' ' + k + terminator
		return ret
	# end stat



	# Optimise code of the form: [load_const]+ [build_tuple|build_list]
	# by computing the result _now_, and replacing the mentioned
	# constants in the coList by this value.
	# The result is sliced for lists, but not for tuples, thus behaving
	# different for this code:
	#
	#	def rett(): return (1,2,3)
	#	x = rett()
	#	print "Equal tuples:", id(x) == id(rett())
	#
	# Returns the replacing code.
	def recodeTupList(self, ltup):
		isList = ltup[-1][0] == 'BUILD_LIST'
		if len(ltup) < 3:	# There's nothing to win for small objects
			return ltup

		# Create the tuple, first as a list
		newconst = []
		for offset in range(len(ltup) - 1):
			junk, ci = ltup[offset]
			newconst.append(self.the_consts.data[ci])
		if isList:
			self.did('load constant list')
		else:
			self.did('load constant tuple')
			newconst = tuple(newconst)

		ret = [ ('LOAD_CONST', self.the_consts.findVal(newconst)) ]
		# Only append a slice for lists
		if isList: ret.append(('SLICE', None))
		return ret
	# end recodeTupList


	# Create a new member 'consts' as a list
	# containing the (revised) constants
	# Optimise all code objects found.
	def optimCodeConsts(self):
		myconsts = []
		for c in self.oldcode.co_consts:
			if type(c) == types.CodeType:
				# optimise once since we'll get
				# our share recursively
				c = optimCode(c, self.strip_setlineno,
					      self.stripdoc, 1, self)
			myconsts.append(c)
		self.the_consts.data = myconsts
		if len(self.oldcode.co_consts) <> len(myconsts):
			self.did('changed co_const size from %d to %d' % (
				len(self.oldcode.co_consts), len(myconsts)))
	# end optimCodeConsts


	def Blocks(self):
		'Divide the code in basic blocks and optimise those'
		blocks = ltup.dividebb(self.oldcode.co_code)
		self.oldi2b = lensum(blocks)
		blocks = map(ltup.str2ltup, blocks)
		blocks = mapj2block(self.oldi2b, blocks)
		for i in range(len(blocks)):
			blocks[i] = optimBlock(self, blocks[i])
		self.blocks = blocks
	# end Blocks


	def jumpifConst(self):
		'Replace conditional jumps by unconditionals if possible'
		for b in self.blocks:
			lc = len(b) - 2	# Presumable position of LOAD_CONST
			# If there is a load_const followed by a conditional jump
			if lc >= 0 and b[lc][0] == 'LOAD_CONST' and b[lc+1][0] in cnd_jump:
				constno = b[lc][1]
				# 'Compute' logical value now:
				const = self.the_consts.data[constno]
				if b[lc+1][0] == 'JUMP_IF_FALSE':
					jump = not const
				else:
					jump = const
				if jump:
					# Replace jump instruction with jump
					newi = ('JUMP_ABSOLUTE', b[lc+1][1])
					b[lc+1] = newi
					self.did('load_const x, conditional jump -> unconditional jump')
					# Don't delete dead code; should
					# be deleted together with the matching POP_TOP
					#del b[lc]
					self.did('delete dead instructions')
				else:
					# Don't jump, just discard the jump instruction
					del b[lc+1:]
					self.did('delete dead instructions', 2)
				# end if
			# end if
		# end for
	# end jumpifConst


	# Compute (optimised-constants-tuple, old-to-new-mapping-list)
	def newConsts(self):
		consts = self.the_consts
		names = self.the_names
		# See which consts are actually used
		for b in self.blocks:
			for op, arg in b:
				if op == 'LOAD_CONST':
					consts.touch(arg)
				elif op in name_instructions:
					names.touch(arg)
		consts.squeeze()
		names.squeeze()
	# end newConsts


	def revJumps(self):
		'''Perform jump-optimisation.
		Revise all arguments of jump instructions.
		They used to point at blocks, now they should get their
		original meaning: relative or absolute byte offsets'''

		def finaltarget(blocks, dest):
			while len(blocks[dest]) == 1 and blocks[dest][0][0] in ('JUMP_FORWARD', 'JUMP_ABSOLUTE'):
				dest = blocks[dest][0][1]
			return dest

		# Optimize jump-to-jump, jump-to-next
		next = 0
		for b in self.blocks:
			next = next + 1
			if len(b):
				op, arg = b[-1]
				if op in rel_jump + abs_jump:
					if arg == next:
						#print 'next =', next, 'op =', op
						if op in cnd_jump:
							b[-1] = ('POP_TOP', None)
							self.did('delete conditional jump to next')
						else:
							del b[-1:]
							self.did('delete jump to next')
					else:
						dest = finaltarget(self.blocks, arg)
						# avoid relative jumps backward
						#if dest <> arg and (op in abs_jump or dest > arg):
						if dest <> arg and op in ('JUMP_FORWARD', 'JUMP_ABSOLUTE'):
							if dest < arg: op = 'JUMP_ABSOLUTE'
							b[-1] = (op, dest)
							self.did('jump over jump')
		# end for

		self.blocks = mapj2offset(self.blocks, self.the_consts.old2new,
			self.the_names.old2new, ltup.revdict(lensum(self.blocks)))
	# end revJumps


	def packCode(self):
		mycode = ''
		for b in self.blocks:
			    mycode = mycode + ltup.ltup2str(b)
		return mycode
	# end packCode

# end Optimising


# Return map of instruction-offsets to block numbers
# for the first instruction of each block.
# 'blocks' is a list of instruction bytestrings or
# instruction tuples
def lensum(blocks):
	ret = {}
	cnt = total = 0
	for b in blocks:
		ret[total] = cnt
		if type(b) == type(''):
			total = total + len(b)
		elif type(b) == type([]):
			total = total + ltup.ltuplen(b)
		else:
			crash
		cnt = cnt + 1
	return ret
# end lensum


def mapj2block(i2b, blocks):
	'''Replace the arguments of all jump instructions by a destination
	block number (i2b is a map of instruction-offsets to blocknumbers).
	Cannot be optimized to only check the last instruction of a basic
	block, because of SETUP_LOOP stuff, which references jumps without
	actually jumping (i.e, it is in rel_jump, but not in cnd_jump or
	unc_jump)'''
	ret = []

	offset = 0
	for blk in blocks:
		blockoffset = 0
		while blockoffset < len(blk):
			op, arg = blk[blockoffset]
			#print op, arg
			if arg == None: offset = offset + 1
			else: offset = offset + 3

			if op in rel_jump:
				target = offset + arg
			elif op in abs_jump:
				target = arg
			else:
				target = None

			if target <> None:
				if i2b.has_key(target):
					target = i2b[target]
				else:
					print 'Unknown label', op, target, 'at', offset
					print 'labels are', i2b#.keys()
					sys.exit(1)
				# Fix code
				blk[blockoffset] = (op, target)
			blockoffset = blockoffset + 1
		# end while
		ret.append(blk)
	# end for
	return ret
# end mapj2block


def mapj2offset(blocks, coMap, nmMap, mp):
	'''Correct the arguments of all jump instructions (inverse of
	mapj2block), correct the arguments for LOAD_CONST to the new indices.'''
	ret = []

	offset = 0
	for ltup in blocks:
		blockoffset = 0
		while blockoffset < len(ltup):
			op, arg = ltup[blockoffset]
			if arg == None: offset = offset + 1
			else: offset = offset + 3

			target = None
			if op == 'LOAD_CONST':
				if arg < 0 or arg >= len(coMap):
					print 'Impossible LOAD_CONST arg:', `arg`, 'len(coMap) =', len(coMap)
				arg = coMap[arg]
				if arg < 0:
					print 'Using unused constant'
					sys.exit(1)
				ltup[blockoffset] = (op, arg)
			elif op in rel_jump:
				target = mp[arg] - offset
			elif op in abs_jump:
				target = mp[arg]
			elif op in name_instructions:
				if arg < 0 or arg >= len(nmMap):
					print 'Impossible', op, 'arg:', `arg`, 'len(nmMap) =', len(nmMap)
				arg = nmMap[arg]
				if arg < 0:
					print 'Using unused name'
					sys.exit(1)
				ltup[blockoffset] = (op, arg)

			if target <> None:
				# Fix jump instruction
				ltup[blockoffset] = (op, target)
			blockoffset = blockoffset + 1
		# end while
		ret.append(ltup)
	# end for
	return ret
# end mapj2offset


# Dictionaries are loaded as follows:
# BUILD_MAP(0), DUP_TOP, LOAD_CONST('value'), ROT_TWO, LOAD_CONST('key'),
# STORE_SUBSTR.  Note the aliasing assumed by STORE_SUBSTR introduced by
# the preceding DUP_TOP.
# Recognize this structure, ignoring SET_LINENO's
# Return the key and value and the offset of the instruction following
# the block, or None if the structure wasn't recognized.
#def namevalnext(block, offset):
#	lenb = len(block)
#	while offset < lenb and block[offset][0] == 'SET_LINENO':
#		offset = offset + 1
#	if offset >= lenb or block[offset][0] <> 'DUP_TOP':
#		print 'No DUP_TOP'
#		return None
#	offset = nextinst(block, offset)
#	if offset >= lenb or block[offset][0] <> 'LOAD_CONST':
#		print 'No LOAD_CONST 1, but', block[offset][0]
#		return None
#	val = block[offset][1]
#	offset = nextinst(block, offset)
#	if block[offset][0] <> 'ROT_TWO':
#		print 'No ROT_TWO'
#		return None
#	offset = nextinst(block, offset)
#	if block[offset][0] <> 'LOAD_CONST':
#		print 'No LOAD_CONST 2, but', block[offset][0]
#		return None
#	name = block[offset][1]
#	offset = nextinst(block, offset)
#	if block[offset][0] <> 'STORE_SUBSCR':
#		print 'No STORE_SUBSCR'
#		return None
#	return name, val, nextinst(block, offset)
## end namevalnext
#
#
#def recodeDict(block, offset, coList):
#	print 'recodeDict: offset =', offset
#	next = nextinst(block, offset)
#	dict = {}
#	nameval = namevalnext(block, next)
#	while nameval:
#		name, val, next = nameval
#		dict[coList[name]] = coList[val]
#		nameval = namevalnext(block, next)
#	if len(dict) > 0:
#		block[offset:next] = [
#			('LOAD_CONST', findConst(coList, dict) ),
#			('SLICE', None) ]	# HIRO: cannot be sliced..
## end recodeDict


def optimBin(self, left, right, offset, block):
	'''Optimize the block.  There is a binary operator at offset, and
	perhaps LOAD_CONST instructions at left and right, that are the
	left and right operand.  Modifies block.  Returns the new offset
	to use for the peephole optimizer.'''
	leftop, leftarg = block[left]
	rightop, rightarg = block[right]

	# Note: the next expression could be optimized to:
	# if not (leftop == rightop == 'LOAD_CONST'):
        if leftop <> 'LOAD_CONST' or rightop <> 'LOAD_CONST':
                return offset

	# Lookup optr and args 
	optr = block[offset][0]
	leftarg = self.the_consts.data[leftarg]
	rightarg = self.the_consts.data[rightarg]

	if optr == 'BINARY_ADD':
		#HIRO: should add strings too
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset
		res = leftarg + rightarg
		block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
		del block[right:offset+1]
		self.did('evaluated binary +')
                return offset - 2

	elif optr == 'BINARY_SUBTRACT':
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset
		res = leftarg - rightarg
		block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
		del block[right:offset+1]
		self.did('evaluated binary -')
                return offset - 2 

	elif optr == 'BINARY_LSHIFT':
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset
		res = leftarg << rightarg
		block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
		del block[right:offset+1]
		self.did('evaluated <<')
                return offset - 2

	elif optr == 'BINARY_RSHIFT':
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset 
		res = leftarg >> rightarg
		block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
		del block[right:offset+1]
		self.did('evaluated >>')
                return offset - 2

	elif optr == 'BINARY_DIVIDE':
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset
		try:	# Catch zero division
			res = leftarg / rightarg
			ok = 1
		except ZeroDivisionError:
			res = sys.exc_value
			ok = 0
		if ok:
			block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
			del block[right:offset+1]
		else:
			# HIRO: should probably generated code to raise
			# this exception in the user code.
			# print 'Warning: constant division by zero'
			return offset
		self.did('evaluated /')
                return offset - 2

	elif optr == 'BINARY_MODULO':
		if type(leftarg) not in numericTypes or type(rightarg) not in numericTypes:
			return offset
		try:	# Catch zero division
			res = leftarg % rightarg
			ok = 1
		except ZeroDivisionError:
			res = sys.exc_value
			ok = 0
		if ok:
			block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
			del block[right:offset+1]
		else:
			# HIRO: should probably generated code to raise
			# this exception in the user code.
			print 'Warning: constant modulo by zero'
			return offset
		self.did('evaluated %')
                return offset - 2

	elif optr == 'BINARY_MULTIPLY':
		# Accept numeric constants, or a string and an (int or long)
		if (type(leftarg) in numericTypes and type(rightarg) in numericTypes) \
		or (type(leftarg) in intAndStrTypes and type(rightarg) in intAndStrTypes \
		    and type(leftarg) <> type(rightarg)):
			res = leftarg * rightarg
			block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
			del block[right:offset+1]
			self.did('evaluated binary *')
                        return offset - 2
	elif optr == 'BINARY_POWER':
		# Accept numeric constants
		if (type(leftarg) in numericTypes and type(rightarg) in numericTypes):
			res = leftarg ** rightarg
			block[left] = ('LOAD_CONST', self.the_consts.findVal(res))
			del block[right:offset+1]
			self.did('evaluated binary **')
                        return offset - 2

# end optimBin


def previnst(ltup, offset):
	'Return position of previous instruction or -1, skipping set_lineno'
	if offset < 0: return offset
	offset = offset - 1
	while ltup[offset][0] == 'SET_LINENO':
		offset = offset - 1
	return offset
# end previnst


def nextinst(ltup, offset):
	'Return position of next instruction or -1, skipping set_lineno'
	if offset < 0: return offset
	offset = offset + 1
	while offset < len(ltup) and ltup[offset][0] == 'SET_LINENO':
		offset = offset + 1
	if offset >= len(ltup): offset = -1
	return offset
# end nextinst

def optimLoad(self, block, offset):
	'Optimize loading of a non-constant; offset points the LOAD_*'
	# Note: this prevents optimisation of more than two
	# LOAD_FAST instructions.  To be fixed by doing real
	# data-flow analysis.
	load, arg = block[offset]
	store = {
		'LOAD_FAST':	'STORE_FAST',
		'LOAD_GLOBAL':	'STORE_GLOBAL',
		'LOAD_LOCAL':	'STORE_LOCAL',
		'LOAD_NAME':	'STORE_NAME',
	}[load]
	prev = previnst(block, offset)
	if prev >= 0:
		prevop, prevarg = block[prev]
		if arg == prevarg:
			if prevop == store:
				# Replace store-load by dup-store
				#block[offset] = block[prev]
				#block[prev] = ('DUP_TOP', None)
				del block[offset]
				block[prev:prev] = [('DUP_TOP', None)]
				self.did('store,load => dup,store')
			elif prevop == load:
				# Replace load-load by load-dup
				block[offset] = ('DUP_TOP', None)
				self.did('load,load => load,dup')
	return offset
# end optimLoad


def optimBlock(self, block):
	'''Optimize a basic block.
	Strip the first of adjescent SET_LINENO instructions
	Strip set_lineno instructions that repeat the last line
	Strip code after unconditional jumps
	Replace (STORE_NAME, 'x'), (LOAD_NAME, 'x') with:
	(DUP_TOP), (STORE_NAME, 'x').
	Remove LOAD_CONST, POP_TOP sequences'''

	offset = 0
	lastline = None		# arg of last SET_LINENO instruction
	loadconst_cnt = 0	# # LOAD_CONST instructions in a row

	while offset < len(block):
		op, arg = block[offset]

		if op in unc_jumpr:	# Strip dead code after this jump
			self.did('dead instruction', len(block) - offset - 1)
			block = block[:offset + 1]
		elif op == 'LOAD_CONST':
			loadconst_cnt = loadconst_cnt + 1
		elif op == 'UNARY_POSITIVE':	# A no-op
			self.did('braindead instruction')
			del block[offset]
			offset = offset - 1
		elif op == 'UNARY_NEGATIVE':
			prev = previnst(block, offset)
			if prev >= 0 and block[prev][0] == 'LOAD_CONST':
				const = self.the_consts.data[block[prev][1]]
				if type(const) in numericTypes:
					const = -const
					del block[offset]
					offset = offset - 1
					block[prev] = ('LOAD_CONST', self.the_consts.findVal(const))
					self.did('evaluate unary -')
				del const
		elif op == 'UNARY_CONVERT':
			prev = previnst(block, offset)
			if prev >= 0 and block[prev][0] == 'LOAD_CONST':
				const = self.the_consts.data[block[prev][1]]
				const = repr(const)
				del block[offset]
				offset = offset - 1
				block[prev] = ('LOAD_CONST', self.the_consts.findVal(const))
				self.did('evaluate repr')
				del const
		elif op == 'UNARY_NOT':
			prev = previnst(block, offset)
			if prev >= 0 and block[prev][0] == 'LOAD_CONST':
				const = self.the_consts.data[block[prev][1]]
				const = not const
				del block[offset]
				offset = offset - 1
				block[prev] = ('LOAD_CONST', self.the_consts.findVal(const))
				self.did('evaluate not')
				del const
		elif op == 'UNARY_INVERT':
			prev = previnst(block, offset)
			if prev >= 0 and block[prev][0] == 'LOAD_CONST':
				const = self.the_consts.data[block[prev][1]]
				if type(const) in numericTypes:
					const = ~const
					del block[offset]
					offset = offset - 1
					block[prev] = ('LOAD_CONST', self.the_consts.findVal(const))
					self.did('evaluate ~')
				else:
					pass # HIRO: Could at least warn,
					# or optimize to a RAISE?
				del const
		elif op in ('BINARY_MULTIPLY', 'BINARY_DIVIDE', 'BINARY_MODULO',
			'BINARY_ADD', 'BINARY_SUBTRACT', 'BINARY_LSHIFT',
			 'BINARY_RSHIFT', 'BINARY_SUBSTR', 'BINARY_POWER'):
				right = previnst(block, offset)
				left = previnst(block, right)
				if left >= 0:
					offset = optimBin(self, left, right, offset, block)
		elif op in ('BUILD_TUPLE', 'BUILD_LIST'):
			if loadconst_cnt >= arg:
				# Build that constant now, not runtime
				offset = offset - arg	# Back to first load_const
				beyond = offset+arg+1	# Beyond the BUILD_* instruction
				newcode = self.recodeTupList(block[offset:beyond])
				block[offset:beyond] = newcode
				offset = offset + len(newcode) - 1	# But mind the upcoming increment
				loadconst_cnt = 0
				del beyond, newcode
		# This outputs an illegal SLICE instruction
		#elif op == 'BUILD_MAP':
		#	recodeDict(block, offset, self.consts)
		elif op == 'POP_TOP' and self.stripdoc:
			if offset > 0 and block[offset-1][0] == 'LOAD_CONST':
				del block[offset-1:offset+1]
				offset = offset - 2
				self.did('delete load,pop')
		elif op in ('LOAD_FAST','LOAD_NAME','LOAD_GLOBAL','LOAD_LOCAL'):
			offset = optimLoad(self, block, offset)
		elif op == 'SET_LINENO':
			if self.strip_setlineno:
				del block[offset]
				offset = offset - 1		# Compensate for upcoming increment
				self.did('set_lineno')
			else:
				if offset > 0 and block[offset-1][0] == 'SET_LINENO':
					# Eat previous SET_LINENO
					del block[offset-1]
					offset = offset - 1	# Compensate for upcoming increment
					self.did('set_lineno')
				elif lastline == arg:
					# Eat this redundant arg
					del block[offset]
					offset = offset - 1	# Compensate for upcoming increment
					self.did('set_lineno')
				lastline = arg

		if op <> 'LOAD_CONST': loadconst_cnt = 0
		offset = offset + 1
	# end while
	return block
# end optimBlock


def optimCode(code, strip_setlineno=0, stripdoc=1, passes=2, parentOpt=None, verbose=0):
	'''Return an optimized code-object'''
	#print 'optimizing', code.co_name, 'in', co_filename
	while passes > 0:
		passes = passes - 1
		O = Optimising(code, strip_setlineno, parentOpt)
		O.optimCodeConsts()
		O.Blocks()
		O.jumpifConst()
		O.newConsts()
		O.revJumps()
		if parentOpt:
			parentOpt.statadd(O.statistics)
		elif verbose > 0:
			x = O.stat()
			if x: print x,

		code = new.code(
			code.co_argcount, code.co_nlocals, code.co_flags,
			O.packCode(), O.the_consts.revision,
			O.the_names.revision, code.co_varnames,
			code.co_filename, code.co_name)

	return code
# end optimCode
